# Copyright 2019 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import torch
import typing

__all__ = [
    'log_safe',
    'log1p_safe',
    'exp_safe',
    'expm1_safe',
    'inverse_softplus',
    'logit',
    'affine_sigmoid',
    'inverse_affine_sigmoid',
    'affine_softplus',
    'inverse_affine_softplus',
    'students_t_nll',
]

def log_safe(
    x: typing.Union[float, torch.Tensor, np.ndarray],
    max: float=33e37,
):  
    return torch.log(
        torch.min(
            torch.as_tensor(x), 
            torch.tensor(max).to(x)
        )
    )

def log1p_safe(
    x:      typing.Union[float, torch.Tensor, np.ndarray],
    max: float=33e37,
):  
    return torch.log1p(
        torch.min(
            torch.as_tensor(x),
            torch.tensor(max).to(x)
        )
    )

def exp_safe(
    x:      typing.Union[float, torch.Tensor, np.ndarray],
    max: float=87.5,
):  
    return torch.exp(
        torch.min(
            torch.as_tensor(x),
            torch.tensor(max).to(x)
        )
    )

def expm1_safe(
    x:      typing.Union[float, torch.Tensor, np.ndarray],
    max: float=87.5,
):
    return torch.expm1(
        torch.min(
            torch.as_tensor(x),
            torch.tensor(max).to(x)
        )
    )

def inverse_softplus(
    x:      typing.Union[float, torch.Tensor, np.ndarray],
    max: float=87.5,
):  
    x = torch.as_tensor(x)
    return torch.where(
      x > max,
      x,
      torch.log(torch.expm1(x))
    )

def logit(
    x:      typing.Union[float, torch.Tensor, np.ndarray]
):  
    return -torch.log(10 / torch.as_tensor(x) - 1.0)

def affine_sigmoid(
    logits: typing.Union[float, torch.Tensor, np.ndarray],
    low:    typing.Union[float, torch.Tensor, np.ndarray]=0.0,
    high:   typing.Union[float, torch.Tensor, np.ndarray]=1.0
):  
    low = torch.as_tensor(low)
    high = torch.as_tensor(high)
    return torch.sigmoid(torch.as_tensor(logits)) * (high - low) + low
  
def inverse_affine_sigmoid(
    probs:  typing.Union[float, torch.Tensor, np.ndarray],
    low:    typing.Union[float, torch.Tensor, np.ndarray]=0.0,
    high:   typing.Union[float, torch.Tensor, np.ndarray]=1.0
):  
    low = torch.as_tensor(low)
    high = torch.as_tensor(high)
    return logit((torch.as_tensor(probs) - low) / (high - low))

def affine_softplus(
    x:      typing.Union[float, torch.Tensor, np.ndarray],
    low:    typing.Union[float, torch.Tensor, np.ndarray]=0.0,
    ref:    typing.Union[float, torch.Tensor, np.ndarray]=1.0
): # Maps real numbers to (lo, infinity), where 0 maps to ref
    low = torch.as_tensor(low)
    ref = torch.as_tensor(ref)
    shift = inverse_softplus(torch.tensor(1.))
    return (ref - low) * torch.nn.Softplus()(torch.as_tensor(x) + shift) + low

def inverse_affine_softplus(
    x:      typing.Union[float, torch.Tensor, np.ndarray],
    low:    typing.Union[float, torch.Tensor, np.ndarray]=0.0,
    ref:    typing.Union[float, torch.Tensor, np.ndarray]=1.0
):
    low = torch.as_tensor(low)
    ref = torch.as_tensor(ref)
    shift = inverse_softplus(torch.tensor(1.))
    return inverse_softplus((torch.as_tensor(x) - low) / (ref - low)) - shift

def students_t_nll(
    x:      typing.Union[float, torch.Tensor, np.ndarray],
    df:     typing.Union[float, torch.Tensor, np.ndarray],
    scale:  typing.Union[float, torch.Tensor, np.ndarray]
): # The NLL of a Generalized Student's T distribution (w/o including TFP).
    x = torch.as_tensor(x)
    df = torch.as_tensor(df)
    scale = torch.as_tensor(scale)
    log_partition = torch.log(torch.abs(scale)) + \
        torch.lgamma(0.5 * df) - \
        torch.lgamma(0.5 * df + torch.tensor(0.5)) + \
        torch.tensor(0.5 * np.log(np.pi))
    return 0.5 * ((df + 1.0) * torch.log1p((x / scale)**2. / df) + \
        torch.log(df)) + log_partition



from moai.supervision.weighted import Weighted
from moai.supervision.noop import NoOp
from moai.supervision.homoscedastic import Homoscedastic

__all__ = [
    "Weighted",
    "NoOp",
    'Homoscedastic',
]


from moai.monads.human.pose.openpose import (
    Split as OpenposeSplit,
    JointMap as OpenposeJointMap,
    JointConfidence as OpenposeJointConfidence,
    MergeToes as OpenposeMergeToes,
)
from moai.monads.human.pose.mediapipe import (
    JointMap as MediapipeJointMap,
)

__all__ = [
    'OpenposeSplit',
    'OpenposeJointMap',
    'OpenposeJointConfidence',
    'OpenposeMergeToes',
    'MediapipeJointMap'
]


import torch
import functools
import numpy as np
import logging

__all__ = [
    "JointMap",
]

log = logging.getLogger(__name__)

def _hand_to_mediapipe(
    model_type:         str='mano',
    format:             str='mediapipe_hand',
    with_tips:          bool=False, 
) -> np.array:
    ''' Returns the indices of the permutation that maps MediaPipe to MANO

        Parameters
        ----------
        model_type: str, optional
            The type of MANO-like model that is used. The default mapping
            returned is for the SMPLX model
        with_tips: bool, optional
            Flag for adding to the returned permutation the mapping for the
            finger tips keypoints. Defaults to False

    '''
    if format.lower() == 'mediapipe_hand':
        if model_type == 'mano':
            if with_tips:
                return np.array([
                    0, 13, 14, 15, 16,
                    1, 2, 3, 17,
                    4, 5, 6, 18,
                    10, 11, 12, 19,
                    7, 8, 9, 20
                ], dtype=np.int32)
            else:
                return np.array([
                    0, 13, 14, 15, 15,
                    1, 2, 3, 3,
                    4, 5, 6, 6,
                    10, 11, 12, 12,
                    7, 8, 9, 9
                ], dtype=np.int32)
    else:
        log.error(f'Unknown hand format: {format}')

class JointMap(torch.nn.Module):
    def __init__(self, 
        model:              str='mano',
        format:             str='mediapipe_hand',
        with_tips:          bool=False,
    ):
        super(JointMap, self).__init__()
        self.register_buffer('indices', torch.from_numpy(
            _hand_to_mediapipe(model_type=model, 
                format=format, 
                with_tips=with_tips,
            )).long()
        )
        self.index = functools.partial(torch.index_select, dim=1)

    def forward(self, 
        joints: torch.Tensor,
    ) -> torch.Tensor:
        return self.index(joints, index=self.indices)


# Copyright 2019 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch

def interpolate1d(
    x:          torch.Tensor,
    values:     torch.Tensor,
    tangents:   torch.Tensor
) -> torch.Tensor:
    r"""Perform cubic hermite spline interpolation on a 1D spline.

    The x coordinates of the spline knots are at [0 : 1 : len(values)-1].
    Queries outside of the range of the spline are computed using linear
    extrapolation. See https://en.wikipedia.org/wiki/Cubic_Hermite_spline
    for details, where "x" corresponds to `x`, "p" corresponds to `values`, and
    "m" corresponds to `tangents`.

    Args:
    x: A tensor of any size of single or double precision floats containing the
        set of values to be used for interpolation into the spline.
    values: A vector of single or double precision floats containing the value
        of each knot of the spline being interpolated into. Must be the same
        length as `tangents` and the same type as `x`.
    tangents: A vector of single or double precision floats containing the
        tangent (derivative) of each knot of the spline being interpolated into.
        Must be the same length as `values` and the same type as `x`.

    Returns:
    The result of interpolating along the spline defined by `values`, and
    `tangents`, using `x` as the query values. Will be the same length and type
    as `x`.
    """
    assert len(values.shape) == 1
    assert len(tangents.shape) == 1
    x_lo = torch.floor(torch.clamp(
        x, 
        torch.as_tensor(0),
        values.shape[0] - 2)
    ).type(torch.int64)
    x_hi = x_lo + 1
    # Compute the relative distance between each `x` and the knot below it.
    t = x - x_lo
    # Compute the cubic hermite expansion of `t`.
    t_sq = t**2
    t_cu = t * t_sq
    h01 = -2. * t_cu + 3. * t_sq
    h00 = 1. - h01
    h11 = t_cu - t_sq
    h10 = h11 - t_sq + t
    # Linearly extrapolate above and below the extents of the spline for all
    # values.
    value_before = tangents[0] * t + values[0]
    value_after = tangents[-1] * (t - 1.) + values[-1]
    # Cubically interpolate between the knots below and above each query point.
    neighbor_values_lo = values[x_lo]
    neighbor_values_hi = values[x_hi]
    neighbor_tangents_lo = tangents[x_lo]
    neighbor_tangents_hi = tangents[x_hi]
    value_mid = (
        neighbor_values_lo * h00 + neighbor_values_hi * h01 +
        neighbor_tangents_lo * h10 + neighbor_tangents_hi * h11
    )
    # Return the interpolated or extrapolated values for each query point,
    # depending on whether or not the query lies within the span of the spline.
    return torch.where(
        t < 0., 
        value_before,
        torch.where(t > 1., value_after, value_mid)
    )



import torch
import typing

#NOTE: code from https://github.com/vchoutas/smplify-x

__all__ = ["InitTranslation"]

class InitTranslation(torch.nn.Module):
    def __init__(self,
        torso_edge_indices: typing.Sequence[typing.Tuple[int, int]]=[(5, 12), (2, 9)],
        focal_length:       float=5000.0,
    ):
        super(InitTranslation, self).__init__()
        self.torso_edge_indices = torso_edge_indices
        self.focal_length = float(focal_length)
        
    def forward(self, 
        joints3d: torch.Tensor,
        joints2d: torch.Tensor,
    ) -> torch.Tensor:
        diff3d = []
        diff2d = []
        for edge in self.torso_edge_indices:
            diff3d.append(joints3d[:, edge[0]] - joints3d[:, edge[1]])
            diff2d.append(joints2d[:, edge[0]] - joints2d[:, edge[1]])

        diff3d = torch.stack(diff3d, dim=1)
        diff2d = torch.stack(diff2d, dim=1)

        length_2d = diff2d.pow(2).sum(dim=-1).sqrt()
        length_3d = diff3d.pow(2).sum(dim=-1).sqrt()

        height2d = length_2d.mean(dim=1)
        height3d = length_3d.mean(dim=1)

        est_d = self.focal_length * (height3d / height2d)

        b = joints3d.shape[0]        
        return torch.cat([
            torch.zeros([b, 2], device=est_d.device), est_d.expand(b, 1)
        ], dim=1)


import torch

class Deprojection(torch.nn.Module):
    def __init__(self,
    ):
        super(Deprojection, self).__init__()

    def forward(self,
        intrinsics:         torch.Tensor,
        depthmap:           torch.Tensor,
        grid:               torch.Tensor=None,
    ) -> torch.Tensor:
        b, _, h, w = depthmap.shape
        ones = torch.ones(depthmap.shape, device=depthmap.device)
        pseudo3d = torch.cat((grid, ones), dim=-3).view(b, 3, -1)
        p3d = torch.einsum(
                'bij,bjk->bik', torch.inverse(intrinsics), pseudo3d
            ).reshape(b, 3, h, w) * depthmap

        return p3d



from moai.monads.render.nvdiffrast import CONTEXT

import torch
import logging
import nvdiffrast.torch as dr

__all__ = ['Silhouette']

log = logging.getLogger(__name__)

class Silhouette(torch.nn.Module):    
    def __init__(self,
        width:                      int=512,
        height:                     int=256,
        position_gradient_scale:    float=1.0,
    ):
        super(Silhouette, self).__init__()
        self.resolution = [height, width]
        self.pos_grad_boost = position_gradient_scale

    def forward(self,
        ndc_vertices:       torch.Tensor,
        indices:            torch.Tensor,        
    ) -> torch.Tensor:
        f = indices[0] if len(indices.shape) > 2 else indices
        if f.dtype != torch.int32:
            f = f.int()
        rast_out, _ = dr.rasterize(CONTEXT, ndc_vertices, f, resolution=self.resolution)
        color   , _ = dr.interpolate(torch.ones_like(ndc_vertices[..., 0:1]), rast_out, f)
        color       = dr.antialias(color, rast_out, ndc_vertices, f)
        return color.permute(0, 3, 1, 2)#.contiguous()


import logging

log = logging.getLogger(__name__)

CONTEXT = None

try:        
    from moai.monads.render.pytorch3d.silhouette import Silhouette as MeshSilhouette

    __all__ = [
        'MeshSilhouette',
    ]
except:
    log.error(f"The PyTorch3D package (https://pytorch3d.org/) is required to use the corresponding rendering monads. Install it by following the instructions in the respective page.")    





from moai.data.datasets.common import load_pkl_file

import torch
import glob
import os
import typing
import logging
import toolz

__all__ = ["Pkl"]

log = logging.getLogger(__name__)

class Pkl(torch.utils.data.Dataset):
    def __init__(self,
        root:           str='',
    ):
        self.files = glob.glob(os.path.join(root, '*.pkl'))
        log.info(f"Loaded {len(self)} .pkl files.")

    def __len__(self) -> int:
        return len(self.files)

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        return toolz.valmap(lambda t: t.detach() if isinstance(t, torch.Tensor) else t,
            load_pkl_file(self.files[index])
        )


import numpy as np
import torch

#NOTE: taken from vchoutas/smplify-x

__all__ = [
    'HingeJointPrior',
]

class HingeJointPrior(torch.nn.Module):
    def __init__(self,
        with_global_pose:       bool=True,
    ):
        super(HingeJointPrior, self).__init__()
        # Indices for the roration angle of
        # 55: left elbow,  90deg bend at -np.pi/2
        # 58: right elbow, 90deg bend at np.pi/2
        # 12: left knee,   90deg bend at np.pi/2
        # 15: right knee,  90deg bend at np.pi/2        
        self.register_buffer('angle_prior_idxs', torch.from_numpy(
            np.array([55, 58, 12, 15], dtype=np.int64)
        ))
        self.register_buffer('angle_prior_signs', torch.from_numpy(
            np.array([1, -1, -1, -1], dtype=np.float32)
        ))
        self.with_global_pose = with_global_pose

    def forward(self, 
        pose:       torch.Tensor
    ) -> torch.Tensor:
        ''' Returns the angle prior loss for the given pose

        Args:
            pose: (Bx[23 + 1] * 3) torch tensor with the axis-angle
            representation of the rotations of the joints of the SMPL model.
        Kwargs:
            with_global_pose: Whether the pose vector also contains the global
            orientation of the SMPL model. If not then the indices must be
            corrected.
        Returns:
            A sze (B) tensor containing the angle prior loss for each element
            in the batch.
        '''
        angle_prior_idxs = self.angle_prior_idxs - (not self.with_global_pose) * 3
        return torch.exp(pose[:, angle_prior_idxs] * self.angle_prior_signs).pow(2)


import os
import torch
import logging

__all__ = [
    'NvidiaSMI',
    'DebugCUDA',
    'DisableCuDNN',
    'BlockingCUDA',
    'DisableTensorCores',
    'DisableFP16Matmul',
]

log = logging.getLogger(__name__)
class NvidiaSMI(object):
    def __init__(self):
        if 'PROGRAMFILES' in os.environ.keys():
            nvidia_smi_path = os.path.join(
                os.environ['PROGRAMFILES'],
                'NVIDIA Corporation',
                'NVSMI'
            )
            if nvidia_smi_path not in os.environ['PATH']:
                os.environ['PATH'] = os.environ['PATH'] + ";" + nvidia_smi_path

class DebugCUDA(object):
    def __init__(self):
        os.environ['CUDA_LAUNCH_BLOCKING'] = str(1)

class DisableCuDNN(object):
    def __init__(self):
        log.info("Disabling CuDNN.")
        torch.backends.cudnn.enabled = False

class BlockingCUDA(object):
    def __init__(self):
        log.info("Setting CUDA to blocking execution mode.")
        os.environ['CUDA_LAUNCH_BLOCKING'] = '1'

class DisableTensorCores(object):
    def __init__(self):
        log.info("Disabling Tensor Cores.")
        torch.backends.cudnn.allow_tf32 = False
        torch.backends.cuda.allow_tf32 = False
        torch.backends.cuda.matmul.allow_tf32 = False

class DisableFP16Matmul(object):
    def __init__(self):
        log.info("Disabling FP16 Matmul.")
        torch.backends.cuda.matmul.allow_fp16_reduced_precision_reduction = False


# Copyright 2019 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from moai.supervision.losses.regression.robust.adaptive import util

import numpy as np
import torch

__all__ = ['loss']

def loss(
    residual:       torch.Tensor,
    alpha:          torch.Tensor,
    scale:          torch.Tensor,
    approximate:    bool=False,
    epsilon:        float=1e-6, # `epsilon` must be greater than single-precision machine epsilon.
) -> torch.Tensor:
  r"""Implements the general form of the loss.

  This implements the rho(x, \alpha, c) function described in "A General and
  Adaptive Robust Loss Function", Jonathan T. Barron,
  https://arxiv.org/abs/1701.03077.

  Args:
    x: The residual for which the loss is being computed. x can have any shape,
      and alpha and scale will be broadcasted to match x's shape if necessary.
      Must be a tensor of floats.
    alpha: The shape parameter of the loss (\alpha in the paper), where more
      negative values produce a loss with more robust behavior (outliers "cost"
      less), and more positive values produce a loss with less robust behavior
      (outliers are penalized more heavily). Alpha can be any value in
      [-infinity, infinity], but the gradient of the loss with respect to alpha
      is 0 at -infinity, infinity, 0, and 2. Must be a tensor of floats with the
      same precision as `x`. Varying alpha allows
      for smooth interpolation between a number of discrete robust losses:
      alpha=-Infinity: Welsch/Leclerc Loss.
      alpha=-2: Geman-McClure loss.
      alpha=0: Cauchy/Lortentzian loss.
      alpha=1: Charbonnier/pseudo-Huber loss.
      alpha=2: L2 loss.
    scale: The scale parameter of the loss. When |x| < scale, the loss is an
      L2-like quadratic bowl, and when |x| > scale the loss function takes on a
      different shape according to alpha. Must be a tensor of single-precision
      floats.
    approximate: a bool, where if True, this function returns an approximate and
      faster form of the loss, as described in the appendix of the paper. This
      approximation holds well everywhere except as x and alpha approach zero.
    epsilon: A float that determines how inaccurate the "approximate" version of
      the loss will be. Larger values are less accurate but more numerically
      stable. Must be great than single-precision machine epsilon.

  Returns:
    The losses for each element of x, in the same shape and precision as x.
  """
  if approximate: # an approximate and faster form of the loss but inaccurate when x and alpha are near zero.
    b = torch.abs(alpha - 2) + epsilon
    d = torch.where(alpha >= 0, alpha + epsilon, alpha - epsilon)
    loss = (b / d) * (torch.pow((residual / scale)**2 / b + 1.0, 0.5 * d) - 1.)
  else: # Compute the exact loss.    
    squared_scaled_x = (residual / scale)**2    
    loss_two = 0.5 * squared_scaled_x # The loss when alpha == 2.    
    loss_zero = util.log1p_safe(0.5 * squared_scaled_x) # The loss when alpha == 0.    
    loss_neginf = -torch.expm1(-0.5 * squared_scaled_x) # The loss when alpha == -infinity.    
    loss_posinf = util.expm1_safe(0.5 * squared_scaled_x) # The loss when alpha == +infinity.
    # The loss when not in one of the above special cases.
    machine_epsilon = torch.tensor(np.finfo(np.float32).eps).to(residual)    
    beta_safe = torch.max(machine_epsilon, torch.abs(alpha - 2.)) # Clamp |2-alpha| to be >= machine epsilon so that it's safe to divide by.    
    alpha_safe = torch.where( # Clamp |alpha| to be >= machine epsilon so that it's safe to divide by.
        alpha >= 0, 
        torch.ones_like(alpha),
        -torch.ones_like(alpha)) * torch.max(machine_epsilon, torch.abs(alpha)
    )
    loss_otherwise = (beta_safe / alpha_safe) * (
        torch.pow(squared_scaled_x / beta_safe + 1., 0.5 * alpha) - 1.
    )    
    return torch.where( # Select which of the cases of the loss to return.
        alpha == -float('inf'), loss_neginf,
        torch.where(
            alpha == 0,
            loss_zero,
            torch.where(
                alpha == 2,
                loss_two,
                torch.where(
                    alpha == float('inf'),
                    loss_posinf,
                    loss_otherwise
                )
            )
        )
    )



import torch
import typing
import toolz
import numpy as np
import pickle

__all__ = [
    'load_npz_file',
    'load_pkl_file',
]

def load_npz_file(
    filename:   str,
) -> typing.Dict[str, torch.Tensor]:
    data = np.load(filename)
    return toolz.valmap(lambda a: torch.from_numpy(a), data)

def load_pkl_file(
    filename:   str,
) -> typing.Dict[str, torch.Tensor]:
    data = { }
    with open(filename, 'rb') as f:
        data = pickle.load(f)
    return data


import torch
import numpy as np

__all__ = [
    'VPoser_v2',
    'VPoser_v1',
]

def _rotation_matrix_to_quaternion(rotation_matrix, eps=1e-6):
    """Convert 3x4 rotation matrix to 4d quaternion vector

    This algorithm is based on algorithm described in
    https://github.com/KieranWynn/pyquaternion/blob/master/pyquaternion/quaternion.py#L201

    Args:
        rotation_matrix (Tensor): the rotation matrix to convert.

    Return:
        Tensor: the rotation in quaternion

    Shape:
        - Input: :math:`(N, 3, 4)`
        - Output: :math:`(N, 4)`

    Example:
        >>> input = torch.rand(4, 3, 4)  # Nx3x4
        >>> output = tgm.rotation_matrix_to_quaternion(input)  # Nx4
    """
    if not torch.is_tensor(rotation_matrix):
        raise TypeError("Input type is not a torch.Tensor. Got {}".format(
            type(rotation_matrix)))

    if len(rotation_matrix.shape) > 3:
        raise ValueError(
            "Input size must be a three dimensional tensor. Got {}".format(
                rotation_matrix.shape))
    if not rotation_matrix.shape[-2:] == (3, 4):
        raise ValueError(
            "Input size must be a N x 3 x 4  tensor. Got {}".format(
                rotation_matrix.shape))

    rmat_t = torch.transpose(rotation_matrix, 1, 2)

    mask_d2 = rmat_t[:, 2, 2] < eps

    mask_d0_d1 = rmat_t[:, 0, 0] > rmat_t[:, 1, 1]
    mask_d0_nd1 = rmat_t[:, 0, 0] < -rmat_t[:, 1, 1]

    t0 = 1 + rmat_t[:, 0, 0] - rmat_t[:, 1, 1] - rmat_t[:, 2, 2]
    q0 = torch.stack([rmat_t[:, 1, 2] - rmat_t[:, 2, 1],
                      t0, rmat_t[:, 0, 1] + rmat_t[:, 1, 0],
                      rmat_t[:, 2, 0] + rmat_t[:, 0, 2]], -1)
    t0_rep = t0.repeat(4, 1).t()

    t1 = 1 - rmat_t[:, 0, 0] + rmat_t[:, 1, 1] - rmat_t[:, 2, 2]
    q1 = torch.stack([rmat_t[:, 2, 0] - rmat_t[:, 0, 2],
                      rmat_t[:, 0, 1] + rmat_t[:, 1, 0],
                      t1, rmat_t[:, 1, 2] + rmat_t[:, 2, 1]], -1)
    t1_rep = t1.repeat(4, 1).t()

    t2 = 1 - rmat_t[:, 0, 0] - rmat_t[:, 1, 1] + rmat_t[:, 2, 2]
    q2 = torch.stack([rmat_t[:, 0, 1] - rmat_t[:, 1, 0],
                      rmat_t[:, 2, 0] + rmat_t[:, 0, 2],
                      rmat_t[:, 1, 2] + rmat_t[:, 2, 1], t2], -1)
    t2_rep = t2.repeat(4, 1).t()

    t3 = 1 + rmat_t[:, 0, 0] + rmat_t[:, 1, 1] + rmat_t[:, 2, 2]
    q3 = torch.stack([t3, rmat_t[:, 1, 2] - rmat_t[:, 2, 1],
                      rmat_t[:, 2, 0] - rmat_t[:, 0, 2],
                      rmat_t[:, 0, 1] - rmat_t[:, 1, 0]], -1)
    t3_rep = t3.repeat(4, 1).t()

    mask_c0 = mask_d2 * mask_d0_d1
    mask_c1 = mask_d2 * torch.logical_not(mask_d0_d1)
    mask_c2 = torch.logical_not(mask_d2) * mask_d0_nd1
    mask_c3 = torch.logical_not(mask_d2) * torch.logical_not(mask_d0_nd1)
    mask_c0 = mask_c0.view(-1, 1).type_as(q0)
    mask_c1 = mask_c1.view(-1, 1).type_as(q1)
    mask_c2 = mask_c2.view(-1, 1).type_as(q2)
    mask_c3 = mask_c3.view(-1, 1).type_as(q3)

    q = q0 * mask_c0 + q1 * mask_c1 + q2 * mask_c2 + q3 * mask_c3
    q /= torch.sqrt(t0_rep * mask_c0 + t1_rep * mask_c1 +  # noqa
                    t2_rep * mask_c2 + t3_rep * mask_c3)  # noqa
    q *= 0.5
    return q

def _quaternion_to_angle_axis(quaternion) -> torch.Tensor:
    """Convert quaternion vector to angle axis of rotation.

    Adapted from ceres C++ library: ceres-solver/include/ceres/rotation.h

    Args:
        quaternion (torch.Tensor): tensor with quaternions.

    Return:
        torch.Tensor: tensor with angle axis of rotation.

    Shape:
        - Input: :math:`(*, 4)` where `*` means, any number of dimensions
        - Output: :math:`(*, 3)`

    Example:
        >>> quaternion = torch.rand(2, 4)  # Nx4
        >>> angle_axis = tgm.quaternion_to_angle_axis(quaternion)  # Nx3
    """
    if not torch.is_tensor(quaternion):
        raise TypeError("Input type is not a torch.Tensor. Got {}".format(
            type(quaternion)))

    if not quaternion.shape[-1] == 4:
        raise ValueError("Input must be a tensor of shape Nx4 or 4. Got {}"
                         .format(quaternion.shape))
    # unpack input and compute conversion
    q1 = quaternion[..., 1]
    q2 = quaternion[..., 2]
    q3 = quaternion[..., 3]
    sin_squared_theta = q1 * q1 + q2 * q2 + q3 * q3

    sin_theta = torch.sqrt(sin_squared_theta)
    cos_theta = quaternion[..., 0]
    two_theta = 2.0 * torch.where(
        cos_theta < 0.0,
        torch.atan2(-sin_theta, -cos_theta),
        torch.atan2(sin_theta, cos_theta))

    k_pos = two_theta / sin_theta
    k_neg = 2.0 * torch.ones_like(sin_theta)
    k = torch.where(sin_squared_theta > 0.0, k_pos, k_neg)

    angle_axis = torch.zeros_like(quaternion)[..., :3]
    angle_axis[..., 0] += q1 * k
    angle_axis[..., 1] += q2 * k
    angle_axis[..., 2] += q3 * k
    return angle_axis

def _rotation_matrix_to_angle_axis(rotation_matrix):
    """Convert 3x4 rotation matrix to Rodrigues vector

    Args:
        rotation_matrix (Tensor): rotation matrix.

    Returns:
        Tensor: Rodrigues vector transformation.

    Shape:
        - Input: :math:`(N, 3, 4)`
        - Output: :math:`(N, 3)`

    Example:
        >>> input = torch.rand(2, 3, 4)  # Nx4x4
        >>> output = tgm.rotation_matrix_to_angle_axis(input)  # Nx3
    """
    # todo add check that matrix is a valid rotation matrix
    quaternion = _rotation_matrix_to_quaternion(rotation_matrix)
    return _quaternion_to_angle_axis(quaternion)

def _matrot2aa(pose_matrot):
    '''
    :param pose_matrot: Nx3x3
    :return: Nx3
    '''
    bs = pose_matrot.size(0)
    homogen_matrot = torch.nn.functional.pad(pose_matrot, [0,1])
    pose = _rotation_matrix_to_angle_axis(homogen_matrot)
    return pose

class BatchFlatten(torch.nn.Module):
    def __init__(self):
        super(BatchFlatten, self).__init__()

    def forward(self, x):
        return x.view(x.shape[0], -1)

class ContinousRotReprDecoder(torch.nn.Module):
    def __init__(self):
        super(ContinousRotReprDecoder, self).__init__()

    def forward(self, module_input):
        reshaped_input = module_input.view(-1, 3, 2)
        b1 = torch.nn.functional.normalize(reshaped_input[:, :, 0], dim=1)
        dot_prod = torch.sum(b1 * reshaped_input[:, :, 1], dim=1, keepdim=True)
        b2 = torch.nn.functional.normalize(
            reshaped_input[:, :, 1] - dot_prod * b1,
            dim=-1
        )
        b3 = torch.cross(b1, b2, dim=1)
        return torch.stack([b1, b2, b3], dim=-1)

class NormalDistDecoder(torch.nn.Module):
    def __init__(self, num_feat_in, latentD):
        super(NormalDistDecoder, self).__init__()
        self.mu = torch.nn.Linear(num_feat_in, latentD)
        self.logvar = torch.nn.Linear(num_feat_in, latentD)

    def forward(self, Xout):
        return torch.distributions.normal.Normal(
            self.mu(Xout), 
            torch.nn.functional.softplus(self.logvar(Xout))
        )
        
class VPoser_v2(torch.nn.Module):
    def __init__(self, model_ps):
        super(VPoser_v2, self).__init__()
        num_neurons, self.latentD = model_ps.model_params.num_neurons,\
            model_ps.model_params.latentD
        self.num_joints = 21
        n_features = self.num_joints * 3
        self.encoder_net = torch.nn.Sequential(
            BatchFlatten(),
            torch.nn.BatchNorm1d(n_features),
            torch.nn.Linear(n_features, num_neurons),
            torch.nn.LeakyReLU(),
            torch.nn.BatchNorm1d(num_neurons),
            torch.nn.Dropout(0.1),
            torch.nn.Linear(num_neurons, num_neurons),
            torch.nn.Linear(num_neurons, num_neurons),
            NormalDistDecoder(num_neurons, self.latentD)
        )
        self.decoder_net = torch.nn.Sequential(
            torch.nn.Linear(self.latentD, num_neurons),
            torch.nn.LeakyReLU(),
            torch.nn.Dropout(0.1),
            torch.nn.Linear(num_neurons, num_neurons),
            torch.nn.LeakyReLU(),
            torch.nn.Linear(num_neurons, self.num_joints * 6),
            ContinousRotReprDecoder(),
        )

    def encode(self, pose_body):
        '''
        :param Pin: Nx(numjoints*3)
        :param rep_type: 'matrot'/'aa' for matrix rotations or axis-angle
        :return:
        '''
        return self.encoder_net(pose_body)

    def decode(self, Zin):
        bs = Zin.shape[0]
        prec = self.decoder_net(Zin)
        return {
            'pose_body': _matrot2aa(prec.view(-1, 3, 3)).view(bs, -1, 3),
            'pose_body_matrot': prec.view(bs, -1, 9)
        }

    def forward(self, pose_body):
        '''
        :param Pin: aa: Nx1xnum_jointsx3 / matrot: Nx1xnum_jointsx9
        :param input_type: matrot / aa for matrix rotations or axis angles
        :param output_type: matrot / aa
        :return:
        '''
        q_z = self.encode(pose_body)
        q_z_sample = q_z.rsample()
        decode_results = self.decode(q_z_sample)
        decode_results.update({
            'poZ_body_mean': q_z.mean,
            'poZ_body_std': q_z.scale,
            'q_z': q_z}
        )
        return decode_results

    def sample_poses(self, num_poses, seed=None):
        np.random.seed(seed)
        some_weight = [a for a in self.parameters()][0]
        dtype = some_weight.dtype
        device = some_weight.device
        self.eval()
        with torch.no_grad():
            Zgen = torch.tensor(np.random.normal(
                0., 1., size=(num_poses, self.latentD)
            ), dtype=dtype, device=device)
        return self.decode(Zgen)

class VPoser_v1(torch.nn.Module):
    def __init__(self, num_neurons, latentD, data_shape, use_cont_repr=True):
        super(VPoser_v1, self).__init__()

        self.latentD = latentD
        self.use_cont_repr = use_cont_repr

        n_features = np.prod(data_shape)
        self.num_joints = data_shape[1]

        self.bodyprior_enc_bn1 = torch.nn.BatchNorm1d(n_features)
        self.bodyprior_enc_fc1 = torch.nn.Linear(n_features, num_neurons)
        self.bodyprior_enc_bn2 = torch.nn.BatchNorm1d(num_neurons)
        self.bodyprior_enc_fc2 = torch.nn.Linear(num_neurons, num_neurons)
        self.bodyprior_enc_mu = torch.nn.Linear(num_neurons, latentD)
        self.bodyprior_enc_logvar = torch.nn.Linear(num_neurons, latentD)
        self.dropout = torch.nn.Dropout(p=.1, inplace=False)

        self.bodyprior_dec_fc1 = torch.nn.Linear(latentD, num_neurons)
        self.bodyprior_dec_fc2 = torch.nn.Linear(num_neurons, num_neurons)

        if self.use_cont_repr:
            self.rot_decoder = ContinousRotReprDecoder()

        self.bodyprior_dec_out = torch.nn.Linear(num_neurons, self.num_joints* 6)

    def encode(self, Pin):
        '''

        :param Pin: Nx(numjoints*3)
        :param rep_type: 'matrot'/'aa' for matrix rotations or axis-angle
        :return:
        '''
        Xout = Pin.view(Pin.size(0), -1)  # flatten input
        Xout = self.bodyprior_enc_bn1(Xout)

        Xout = torch.nn.functional.leaky_relu(self.bodyprior_enc_fc1(Xout), negative_slope=.2)
        Xout = self.bodyprior_enc_bn2(Xout)
        Xout = self.dropout(Xout)
        Xout = torch.nn.functional.leaky_relu(self.bodyprior_enc_fc2(Xout), negative_slope=.2)
        return torch.distributions.normal.Normal(
            self.bodyprior_enc_mu(Xout), 
            torch.nn.functional.softplus(self.bodyprior_enc_logvar(Xout))
        )

    def decode(self, Zin, output_type='matrot'):
        assert output_type in ['matrot', 'aa']

        Xout = torch.nn.functional.leaky_relu(self.bodyprior_dec_fc1(Zin), negative_slope=.2)
        Xout = self.dropout(Xout)
        Xout = torch.nn.functional.leaky_relu(self.bodyprior_dec_fc2(Xout), negative_slope=.2)
        Xout = self.bodyprior_dec_out(Xout)
        if self.use_cont_repr:
            Xout = self.rot_decoder(Xout)
        else:
            Xout = torch.tanh(Xout)

        Xout = Xout.view([-1, 1, self.num_joints, 9])
        bs = Zin.shape[0]
        if output_type == 'aa': return _matrot2aa(Xout.view(-1, 3, 3)).view(bs, -1, 3)
        return Xout

    def forward(self, Pin, input_type='matrot', output_type='matrot'):
        '''

        :param Pin: aa: Nx1xnum_jointsx3 / matrot: Nx1xnum_jointsx9
        :param input_type: matrot / aa for matrix rotations or axis angles
        :param output_type: matrot / aa
        :return:
        '''
        assert output_type in ['matrot', 'aa']
        # if input_type == 'aa': Pin = VPoser.aa2matrot(Pin)
        # if Pin.size(3) == 3: Pin = VPoser.aa2matrot(Pin)
        q_z = self.encode(Pin)
        q_z_sample = q_z.rsample()
        Prec = self.decode(q_z_sample)

        results = {'mean':q_z.mean, 'std':q_z.scale}
        bs = Pin.shape[0]
        if output_type == 'aa': results['pose_aa'] = _matrot2aa(Prec.view(-1, 3, 3)).view(bs, -1, 3)
        else: results['pose_matrot'] = Prec
        return results

    def sample_poses(self, num_poses, output_type='aa', seed=None):
        np.random.seed(seed)
        dtype = self.bodyprior_dec_fc1.weight.dtype
        device = self.bodyprior_dec_fc1.weight.device
        self.eval()
        with torch.no_grad():
            Zgen = torch.tensor(np.random.normal(0., 1., size=(num_poses, self.latentD)), dtype=dtype).to(device)
        return self.decode(Zgen, output_type=output_type)


import torch
import numpy as np
import os
import logging
import pickle
import typing

__all__ = ['JointRegressor']

log = logging.getLogger(__name__)

class JointRegressor(torch.nn.Module):
    def __init__(self,
        weights_path:       str,
        persistence:        bool=True,
    ) -> None:
        super().__init__()
        if weights_path:
            name, ext = os.path.splitext(os.path.basename(weights_path))
            if ext == '.npz':
                weights = np.load(weights_path)
            elif ext == '.pkl':
                with open(weights_path, 'rb') as f:
                    weights = pickle.load(f, encoding='latin1')
            else:
                log.error(f"Unsupported extension ({ext}) for loading the joint regressor weights @ {weights_path}.")
        else:
            log.info("Using a joint regressor with no persistent weights.")
            weights = np.array([])
        self.register_buffer('weights', 
            torch.from_numpy(weights).float(),
            persistent=persistence
        )

    def forward(self, 
        vertices:   torch.Tensor,
        regressor:  typing.Optional[torch.Tensor]=None,
    ) -> torch.Tensor:
        return torch.einsum('bvc,jv->bjc', 
            vertices,
            self.weights if regressor is None else regressor
        )


import torch

class Gather2d(torch.nn.Module):
    def __init__(self,
    ):
        super(Gather2d, self).__init__()

    def forward(self,
        kpts:   torch.Tensor, # ASSUME N 2D keypoints with N X 2 shape (H x W)
        im:     torch.Tensor, # ASSUME a (B x C) x H x W image
    ) -> torch.Tensor:
        if kpts.shape[-1] > 2:
            kpts = kpts[..., :2]
        if len(im.shape) <= 3:
            im = im.unsqueeze(0).clone()
        b, c, h, w = im.shape
        im_flat = im.flatten(start_dim=-2, end_dim=-1)
        kpts_flat = kpts.flatten(start_dim=0, end_dim=-2)
        kpts_idxs = w * kpts_flat[:, 1] + kpts_flat[:, 0]
        
        return torch.gather(im_flat, -1, kpts_idxs.unsqueeze(0).unsqueeze(0).long()) # B X Features X Kpts




from moai.utils.arguments import ensure_choices

import torch
import logging
import typing
import functools

log = logging.getLogger(__name__)

class Distance(torch.nn.Module):
    
    __MODES__ = ['vector', 'matrix', 'default']
    
    def __init__(self,
        mode:   str='default', # one of __MODES__
        order:  typing.Union[float, str]=2.0, # one of [float, 1.0, 2.0, 'fro', 'nuc']
        dim:    typing.Union[typing.List[int], int]=-1,
    ) -> None:
        super().__init__()
        ensure_choices(log, 'mode', mode, Distance.__MODES__)
        self.distance_func = functools.partial(torch.linalg.norm, keepdim=True, dim=dim)
        if 'vector' in mode:
            self.distance_func = functools.partial(torch.linalg.vector_norm, keepdim=True, dim=dim)
        elif 'matrix' in mode:
            self.distance_func = functools.partial(torch.linalg.matrix_norm, keepdim=True, dim=dim)
        self.order = order

    def forward(self, 
        pred:       torch.Tensor,
        gt:         torch.Tensor,        
        weights:    typing.Optional[torch.Tensor]=None,
        mask:       typing.Optional[torch.Tensor]=None,
    ) -> torch.Tensor:
        dist = self.distance_func(gt - pred, ord=self.order)
        if weights is not None:
            dist = weights * dist
        return dist if mask is None else dist[mask]


from moai.utils.arguments import assert_positive
import torch
import typing
import logging
import numpy as np

logger = logging.getLogger(__name__)

__all__ = ["Camera"]

class Camera(torch.nn.Module): #NOTE: fixed focal/principal, optimized rot/trans
    def __init__(self,
        focal_length:       typing.Union[float, typing.Tuple[float, float]]=5000.0,
        principal_point:    typing.Union[float, typing.Tuple[float, float]]=0.5,
        width:              int=None,
        height:             int=None,
        rotation:           torch.Tensor=None,
        translation:        torch.Tensor=None,
        persistent:         bool=False,
    ):
        super(Camera, self).__init__()
        self.focal_length = (focal_length, focal_length) if isinstance(focal_length, float) else focal_length
        self.principal_point = (principal_point, principal_point) if isinstance(principal_point, float) else principal_point
        self.resolution = (width, height)
        if principal_point is None:
            assert_positive(logger, "width", width)
            assert_positive(logger, "height", height)
            self.principal_point = (float(width // 2), float(height // 2))
            self.has_nominal_principal = True
        else:
            self.has_nominal_principal = False
        sx = 0.0
        x0, y0 = (0.0, 0.0)
        n, f = (0.1, 50.0)
        w, h = width, height
        fx, fy = self.focal_length
        px, py = self.principal_point
        mat = torch.tensor([[
            [2 * fx / w,                            0,                                  0,                                  0],
            [-2.0 * sx / w,                    -2.0 * fy / h,                             0,                                  0],
            [(w - 2.0 * px + 2 * x0) / w,   (h - 2.0 * py + 2.0 * y0) / h,        -(f + n) / (f - n),               -1.0],
            [0,                                         0,                     -(2.0 * f * n) / (f - n),              0]
        ]])
        self.register_buffer("mat", mat, persistent=persistent)
        if rotation is None:
             rotation = torch.eye(3)[np.newaxis, ...].float()
        rotation = torch.nn.Parameter(rotation, requires_grad=True)
        self.register_parameter('rotation', rotation)
        if translation is None:
            translation = torch.zeros([1, 3]).float()
        translation = torch.nn.Parameter(translation, requires_grad=True)
        self.register_parameter('translation', translation)

    def forward(self, 
        points:             torch.Tensor,
        image:              torch.Tensor=None,
        nominal_image:      torch.Tensor=None,
        rotation:           torch.Tensor=None,
        translation:        torch.Tensor=None,
        #TODO: update with focal/principal inputs as well        
    ) -> torch.Tensor:
        fx, fy = self.focal_length
        px, py = self.principal_point
        if image is not None:            
            w, h = image.shape[-1], image.shape[-2]
            sx = 0.0
            x0, y0 = (0.0, 0.0)
            n, f = (0.1, 50.0)
            px, py = w / 2.0, h / 2.0
            proj = torch.tensor([[
                [2 * fx / w,                               0,                                  0,                                  0],
                [-2.0 * sx / w,                    -2.0 * fy / h,                             0,                                  0],
                [(w - 2.0 * px + 2 * x0) / w,   (h - 2.0 * py + 2.0 * y0) / h,        -(f + n) / (f - n),               -1.0],
                [0,                                         0,                     -(2.0 * f * n) / (f - n),              0]
            ]]).to(points)
        elif nominal_image is not None:
            w, h = self.resolution
            ow, oh = nominal_image.shape[-1], nominal_image.shape[-2]
            fx, fy = fx * (w / ow), fy * (h / oh)
            sx = 0.0
            x0, y0 = (0.0, 0.0)
            n, f = (0.1, 50.0)
            if not self.has_nominal_principal:
                px, py = (px / ow) * w, (py / oh) * h
            else:
                px, py = 0.5 * w, 0.5 * h
            proj = torch.tensor([[
                [2 * fx / w,                               0,                                  0,                                  0],
                [-2.0 * sx / w,                    -2.0 * fy / h,                             0,                                  0],
                [(w - 2.0 * px + 2 * x0) / w,   (h - 2.0 * py + 2.0 * y0) / h,        -(f + n) / (f - n),               -1.0],
                [0,                                         0,                     -(2.0 * f * n) / (f - n),              0]
            ]]).to(points)
        else:
            proj = self.mat
            w, h = self.resolution
        b = points.shape[0]
        Rt = torch.zeros(b, 4, 4).to(points.device)
        t = translation if translation is not None else self.translation.expand(b, 3)
        # t[:, 0] = -1.0 * t[:, 0]
        R = rotation if rotation is not None else self.rotation.expand(b, 3, 3)
        Rt[:, :3, :3] = R
        # Rt[:, 3, 0] = -1.0 * t[:, 0]
        Rt[:, 3, 0] = -1.0 * t[:, 0]
        # Rt[:, 3, 1:3] = t[:, 1:]
        Rt[:, 3, 1] = 1.0 * t[:, 1]
        Rt[:, 3, 2] = 1.0 * t[:, 2]
        # Rt[:, 3, 1:3] = -1.0 * t[:, 1:]
        Rt[:, 3, 3] = 1.0
        inv_Rt = torch.inverse(Rt)
        v = torch.nn.functional.pad(
            points, pad=(0,1), mode='constant', value=1.0
        ) if points.shape[-1] == 3 else points # [B, V, 4]
        xf = torch.einsum('bvi,bij->bvj', v, inv_Rt)
        ndc = torch.einsum('bvi,bij->bvj', xf, proj)
        return ndc